/* wptKeysigDlg.cpp - Key signature listing
 *	Copyright (C) 2001-2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>
#include <time.h>

#include "resource.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptDlgs.h"
#include "wptW32API.h"
#include "wptNLS.h"
#include "wptKeylist.h"
#include "wptKeyserver.h"
#include "wptUTF8.h"
#include "wptTypes.h"
#include "wptVersion.h"
#include "wptErrors.h"
#include "wptKeyEdit.h"

static subclass_s siglist_proc;


static int
is_sig (listview_ctrl_t lv, int pos)
{
    char tmpbuf[200];

    if (pos == -1)
	pos = listview_get_curr_pos (lv);
    if (pos == -1)
	return 0;
    listview_get_item_text (lv, pos, SL_COL_UID, tmpbuf, sizeof (tmpbuf)-1);
    if (*tmpbuf == ' ')
	return -1;
    return 0;
}


/* Delete the selected signature from list view @lv. The
   key is given in @key. */
static int
do_delsig (HWND dlg, listview_ctrl_t lv, winpt_key_t key)
{
    gpgme_error_t err;
    GpgKeyEdit *ke;
    char keyid[32];
    int pos, npos, id, is_selfsig=0;
    int signo=0, uidno=0;

    npos = pos = listview_get_curr_pos (lv);
    if (!is_sig (lv, -1))
	return -1;
    listview_get_item_text (lv, pos, SL_COL_KEYID, keyid, sizeof (keyid)-1);
    if (!strncmp (key->keyid, keyid+2, 8))
	is_selfsig = 1;
    
    while (pos > 0 && is_sig (lv, pos)) {
	signo++;
	pos--;
    }
    pos = npos;
    while (npos > 0) {
	if (!is_sig (lv, npos))
	    uidno++;
	npos--;
    }
    uidno++;

    /* XXX: do not allow to delete the self signature */
    id = log_box (_("Key Manager"), is_selfsig? MB_WARN_ASK : MB_QUEST_ASK,
		  _("Are you really sure you want to delete this %s from\n\n"
		  "  \"%s\""), is_selfsig? _("self signature") : _("signature"),
		  key->uid);
    if (id == IDNO)
	return 0;

    ke = new GpgKeyEdit (key->keyid);
    if (!ke)
	BUG (NULL);    
    err = ke->delUseridSignature (uidno, signo);
    if (err)
	msg_box (dlg, gpgme_strerror (err), _("Key Manager"), MB_ERR);
    else
	listview_del_item (lv, pos);
    delete ke;
    if (!err)
	key->update = 1;
    return err? -1 : 0;
}


/* Dialog box procedure to display the signature properties. */
BOOL CALLBACK
sigprops_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static gpgme_key_sig_t ks;
    gpgme_key_t issuer;
    char tmpbuf[256];
    struct {
	unsigned int exportable:1;
	unsigned int expired:1;
	unsigned int nrev:1;
	unsigned int rev:1;
	const char *alg;
	int _class;
    } ctx;
    const char *fmt_templ = _("%s %s signature");

    switch (msg) {
    case WM_SYSCOMMAND:
	if (LOWORD (wparam) == SC_CLOSE)
            EndDialog (dlg, TRUE);
        return FALSE;

    case WM_INITDIALOG:
	SetWindowText (dlg, _("Signature Properties"));
	SetDlgItemText (dlg, IDC_SIGPROPS_EXP, _("Exportable"));
	SetDlgItemText (dlg, IDC_SIGPROPS_NREV, _("Non-revocably"));
	SetDlgItemText (dlg, IDC_SIGPROPS_EXPIRED, _("Expired"));
	SetDlgItemText (dlg, IDC_SIGPROPS_REV, _("Revoked"));
	SetDlgItemText (dlg, IDC_SIGPROPS_CLASSINF, _("Class"));
	SetDlgItemText (dlg, IDC_SIGPROPS_EXPSTR, _("Expire date"));
	SetDlgItemText (dlg, IDC_SIGPROPS_KEYINF, _("Issuer key"));
	SetDlgItemText (dlg, IDC_SIGPROPS_KEYIDINF, _("Issuer key ID"));
	ks = (gpgme_key_sig_t)lparam;
	if (!ks)
	    BUG (0);
	memset (&ctx, 0, sizeof ctx);
	ctx._class = ks->sig_class;
	if (ctx._class == 0)
	    ctx._class = 16;    
	ctx._class -= 6;
	if (ks->pubkey_algo == GPGME_PK_DSA)
	    ctx.alg = "DSA";
	else if (ks->pubkey_algo == GPGME_PK_RSA)
	    ctx.alg = "RSA";
	else
	    ctx.alg = "ELG";
	ctx.exportable = ks->exportable;
	_snprintf (tmpbuf, DIM (tmpbuf)-1, fmt_templ,
		   ctx.exportable? _("Exportable") : _("Non-exportable"), ctx.alg);
	SetDlgItemText (dlg, IDC_SIGPROPS_INFO, tmpbuf);
	
	SetDlgItemText (dlg, IDC_SIGPROPS_KEYID, ks->keyid+8);
	SetDlgItemInt (dlg, IDC_SIGPROPS_CLASS, ctx._class, FALSE);
	if (ctx.exportable)
	    CheckDlgButton (dlg, IDC_SIGPROPS_EXP, BST_CHECKED);
	if (!get_pubkey (ks->keyid+8, &issuer)) {
	    char *uid = utf8_to_native (issuer->uids->uid);
	    SetDlgItemText (dlg, IDC_SIGPROPS_ISSUER, uid);
	    safe_free (uid);
	}
	else
	    SetDlgItemText (dlg, IDC_SIGPROPS_ISSUER, _("  user ID not found"));
	if (ks->expires == 0) {
	    ShowWindow (GetDlgItem (dlg, IDC_SIGPROPS_EXPSTR), SW_HIDE);
	    ShowWindow (GetDlgItem (dlg, IDC_SIGPROPS_EXPDATE), SW_HIDE);
	}
	else {
	    SYSTEMTIME st;
	    struct tm *tm;
	    time_t t = time (NULL);
	    time_t tmp = ks->expires;

	    tm = localtime (&tmp);
	    memset (&st, 0, sizeof st);
	    st.wYear = tm->tm_year+1900;
	    st.wMonth = tm->tm_mon;
	    st.wDay = tm->tm_mday;
	    DateTime_SetSystemtime (GetDlgItem (dlg, IDC_SIGPROPS_EXPDATE),
				    GDT_VALID, &st);

	    tm = localtime (&t);
	    tm->tm_mon++;
	    tm->tm_year += 1900;
	    if (tm->tm_year > st.wYear)
		ctx.expired = 1;
	    else if (tm->tm_mon > st.wMonth)
		ctx.expired = 1;
	    if (ctx.expired)
		CheckDlgButton (dlg, IDC_SIGPROPS_EXPIRED, BST_CHECKED);
	}
	SetDlgItemText (dlg, IDC_SIGPROPS_EXP, _("Exportable"));
	SetDlgItemText (dlg, IDC_SIGPROPS_NREV, _("Non-revocably"));
	SetDlgItemText (dlg, IDC_SIGPROPS_REV, _("Revoked"));
	SetDlgItemText (dlg, IDC_SIGPROPS_EXPIRED, _("Expired"));
	SetWindowText (dlg, _("Signature Properties"));
	SetForegroundWindow (dlg);
	center_window (dlg, NULL);
	return TRUE;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDOK:
	    EndDialog (dlg, TRUE);
	    break;

	case IDCANCEL:
	    EndDialog (dlg, FALSE);
	    break;
	}
    }

    return FALSE;
}


static BOOL CALLBACK
subclass_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    listview_ctrl_t lv;
    winpt_key_t key;
    gpgme_key_sig_t ks;

    switch (msg) {
    case WM_KEYUP:
        int virt_key = (int)wparam;
	key = (winpt_key_t)siglist_proc.opaque;
	lv = key->callback.ctl;
        if (virt_key == VK_SPACE) {
	    if (is_sig (lv, -1)) {
		ks = (gpgme_key_sig_t)listview_get_item2 (lv, listview_get_curr_pos (lv));
		DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_SIGPROPS, dlg,
				sigprops_dlg_proc, (LPARAM)ks);
	    }
	}
	else if (virt_key == VK_DELETE)
	    do_delsig (dlg, lv, key);
	break;
    }
    return CallWindowProc (siglist_proc.old, dlg, msg, wparam, lparam);
}


/* Return 1 if the list view @lv contains non-available keys. */
static int
check_for_missing_keys (listview_ctrl_t lv)
{
    int i, n;
    char id[128];
    
    n = listview_count_items( lv, 0 );
    for( i = 0; i < n; i++ ) {
        listview_get_item_text (lv, i, SL_COL_VALID, id, sizeof (id) - 1);
        if (!strncmp (id, "NOKEY", 5))
            return 1;
    }
    
    return 0;
}


/* Receive all missing keys marked in the list @lv. */
static int
recv_missing_keys (HWND dlg, listview_ctrl_t lv)
{
    char id[128], keyid[18+1];
    int i, n, cnt=0;

    i = msg_box (dlg, _("Really receive all missing keys?"),
		 _("Key Manager"), MB_YESNO|MB_INFO);
    if (i == IDNO)
	return 0;
    
    n = listview_count_items (lv, 0);
    for (i = 0; i < n; i++) {
        listview_get_item_text (lv, i, SL_COL_VALID, id, sizeof (id) - 1);
        if (!strncmp (id, "NOKEY", 5)) {
            listview_get_item_text (lv, i, SL_COL_KEYID, 
				    keyid, sizeof (keyid) -1);
            if (!hkp_recv_key (dlg, default_keyserver,
			       default_keyserver_port, keyid, 0, 0)) {
		keycache_update (0, keyid);
		cnt++;
	    }
	    else
		break;
        }
    }
    return cnt;
}


/* Create a mini popup with available choices. */
static void
do_create_popup (HWND dlg)
{
    HMENU hm, sm;
    POINT p;

    GetCursorPos (&p);
    hm = LoadMenu (glob_hinst, MAKEINTRESOURCE (IDR_WINPT_KEYSIG_CTX));
    sm = GetSubMenu (hm, 0);
    set_menu_text (sm, ID_SIGCTX_PROPS, _("Signature &Properties"));
    set_menu_text (sm, ID_SIGCTX_KEYPROPS, _("Signing &Key Properties"));
    TrackPopupMenu (sm, TPM_RIGHTALIGN, p.x, p.y, 0, dlg, NULL);
    DestroyMenu (hm);
    DestroyMenu (sm);
}


/* Fetch a single key and extract the keyid from the listview
   entry at the pos @idx. */
static int
recv_single_key (HWND dlg, listview_ctrl_t lv, int idx)
{
    char keyid[32];
    int rc;

    listview_get_item_text (lv, idx, SL_COL_KEYID, keyid, DIM (keyid)-1);	
    rc = hkp_recv_key (dlg, default_keyserver, 
		       default_keyserver_port, keyid, 0, 0);
		
    if (!rc)
	keycache_update (0, keyid);
    return rc;
}


/* Load the key property dialog with the selected key from @lv. */
static void
do_load_keyprops (HWND dlg, listview_ctrl_t lv)
{
    winpt_key_s k;
    gpgme_key_t key;
    char keyid[32] = {0};
    char status[64] = {0}, creation[64] = {0};
    int n = listview_get_curr_pos (lv);

    listview_get_item_text (lv, n, SL_COL_VALID, status, DIM (status)-1);
    listview_get_item_text (lv, n, SL_COL_CREATE, creation, DIM (creation)-1);
    listview_get_item_text (lv, n, SL_COL_KEYID, keyid, DIM (keyid)-1);
    if (!strcmp (status, "NOKEY")) {
	int id = msg_box (dlg, _("Key not found in keyring, do you want to fetch it from the keyserver?"),
			  _("Key Manager"), MB_QUEST_ASK);
	if (id == IDNO)
	    return;
	if (recv_single_key (dlg, lv, n))		
	    return;
    }
    
    if ((strlen (keyid) < 3 ||get_pubkey (keyid, &key))) {
	if (strlen (creation) > 0)
	    msg_box (dlg, _("Key not found in keyring."), _("Key Manager"), MB_INFO);
	return;
    }
    memset (&k, 0, sizeof k);
    k.keyid = keyid;
    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_KEYPROPS, dlg,
		    keyprops_dlg_proc, (LPARAM)&k);
}


/* Dialog box procedure to list signatures. */
BOOL CALLBACK
keysig_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv = NULL;
    static struct winpt_key_s *k;
    gpgme_key_sig_t ks;
    HWND sl;
    char inf[384];
    int idx = 0;    
    
    switch (msg) {
    case WM_INITDIALOG:
        k = (winpt_key_t) lparam;
	if (!k)
	    BUG (0);
        if (k->uid)
            _snprintf (inf, DIM (inf)-1, _("Signature List for \"%s\""), k->uid);
        SetWindowText (dlg, inf);
        SetDlgItemText (dlg, IDC_KEYSIG_RECVKEY, _("&Receive Key"));
	SetDlgItemText (dlg, IDC_KEYSIG_SIGPROPS, _("&Properties"));

	lv = siglist_load (GetDlgItem (dlg, IDC_KEYSIG_LIST), k->keyid);
	if (!check_for_missing_keys (lv))
	    EnableWindow (GetDlgItem (dlg, IDC_KEYSIG_RECVKEY), FALSE);
	EnableWindow (GetDlgItem (dlg, IDC_KEYSIG_SIGPROPS), FALSE);
	k->callback.ctl = lv;
	sl = GetDlgItem (dlg, IDC_KEYSIG_LIST);
	siglist_proc.dlg = dlg;
	siglist_proc.opaque = k;
	siglist_proc.current = (WNDPROC)subclass_dlg_proc;
	siglist_proc.old = (WNDPROC)GetWindowLong (sl, GWL_WNDPROC);
	if (siglist_proc.old) {
	    if (!SetWindowLong (sl, GWL_WNDPROC, (LONG)siglist_proc.current))  {
		msg_box (dlg, "Could not set keylist window procedure.",
		         _("Key Manager"), MB_ERR);
		BUG (0);
	    }
	}
        SetForegroundWindow (dlg);
	center_window (dlg, NULL);
        return TRUE;
        
    case WM_DESTROY:
        if (lv) {
            siglist_delete (lv);
            lv = NULL;
        }
        return FALSE;
        
    case WM_SYSCOMMAND:
        if (LOWORD (wparam) == SC_CLOSE)
            EndDialog (dlg, TRUE);
        return FALSE;

    case WM_NOTIFY:
	NMHDR *notify;
        
        notify = (NMHDR *)lparam;
	if (!notify)
	    return FALSE;
        if (notify->code == NM_DBLCLK 
	    && notify->idFrom == IDC_KEYSIG_LIST)
	    do_load_keyprops (dlg, lv);
	if (notify->code == NM_RCLICK &&
	    notify->idFrom == IDC_KEYSIG_LIST &&
	    is_sig (lv, -1))
	    do_create_popup (dlg);
	if (notify->code == LVN_ITEMCHANGED &&
	    ((LPNMLISTVIEW)lparam)->uNewState) {
	    idx = listview_get_curr_pos (lv);
	    listview_get_item_text (lv, idx, SL_COL_VALID, inf, DIM (inf)-1);
	    EnableWindow (GetDlgItem (dlg, IDC_KEYSIG_RECVKEY),
			  strcmp (inf, "NOKEY") == 0? TRUE: FALSE);
	    EnableWindow (GetDlgItem (dlg, IDC_KEYSIG_SIGPROPS),
			  is_sig (lv, -1)? TRUE : FALSE);
	}
	if (notify->code == LVN_COLUMNCLICK) {
	    NMLISTVIEW *nft = (LPNMLISTVIEW) lparam;
	    int sortby = 0;
	    switch (nft->iSubItem) {
	    //case 0:  sortby = KEY_SORT_USERID; break;
	    case 1: sortby = KEY_SORT_VALIDITY; break;
	    case 2: sortby = SIG_SORT_CLASS; break;
	    case 3: sortby = KEY_SORT_CREATED; break;
	    case 4: sortby = KEY_SORT_KEYID; break;
	    case 5: sortby = SIG_SORT_EXPIRE; break;
	    case 6: sortby = KEY_SORT_ALGO; break;
	    default: return TRUE;
	    }
	    siglist_sort (lv, sortby);
	}
	break;
        
    case WM_COMMAND:
        switch (LOWORD (wparam)) {
	case ID_SIGCTX_KEYPROPS:
	    do_load_keyprops (dlg, lv);
	    break;

	case ID_SIGCTX_PROPS:
	    if (is_sig (lv, -1)) {
		ks = (gpgme_key_sig_t)listview_get_item2 (lv, listview_get_curr_pos (lv));
		DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_SIGPROPS, dlg,
				sigprops_dlg_proc, (LPARAM)ks);
	    }
	    break;

        case IDC_KEYSIG_RECVKEY:
	    idx = listview_get_curr_pos (lv);
            if (idx == -1)
		recv_missing_keys (dlg, lv);
	    else
		recv_single_key (dlg, lv, idx);
            return TRUE;

	case IDC_KEYSIG_SIGPROPS:
	    if (is_sig (lv, -1)) {
		ks = ks = (gpgme_key_sig_t)listview_get_item2 (lv, listview_get_curr_pos (lv));
		DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_SIGPROPS, dlg,
				sigprops_dlg_proc, (LPARAM)ks);
	    }
	    return TRUE;
            
        case IDOK:
            EndDialog (dlg, TRUE);
            return TRUE;
        }
        break;
    }
    return FALSE;
}
